#include "kernel.h"
#include <stdio.h>


// Get a matrix element 
__device__ float GetElement(const Matrix A, int row, int col) 
{   
   return A.elements[row * A.stride + col]; 
}

// Set a matrix element 
__device__ void SetElement(Matrix A, int row, int col, float value) 
{    
  A.elements[row * A.stride + col] = value;
}

// Get the BLOCK_SIZExBLOCK_SIZE sub-matrix Asub of A that is 
// located col sub-matrices to the right and row sub-matrices down 
// from the upper-left corner of A 
__device__ Matrix GetSubMatrix(Matrix A, int row, int col) 
{    
  Matrix Asub;    
  Asub.width    = BLOCK_SIZE;    
  Asub.height   = BLOCK_SIZE;    
  Asub.stride   = A.stride;    
  Asub.elements = &A.elements[A.stride * BLOCK_SIZE * row  + BLOCK_SIZE * col];   
   return Asub;
}

// // Thread block size 
// #define BLOCK_SIZE 16 

// Forward declaration of the matrix multiplication kernel 
__global__ void MatMulKernel(const Matrix, const Matrix, Matrix); 

// Matrix multiplication - Host code 
// Matrix dimensions are assumed to be multiples of BLOCK_SIZE 
void MatMul(const Matrix A, const Matrix B, Matrix C) 
{    
  // Load A and B to device memory    
  // 加载矩阵 A 和 B 到设备内存中

  // 矩阵 A --> 矩阵 d_A
  Matrix d_A;    
  d_A.width = d_A.stride = A.width;
  d_A.height = A.height;    
  size_t size = A.width * A.height * sizeof(float);   
  cudaMalloc(&d_A.elements, size);   
  cudaMemcpy(d_A.elements, A.elements, size, cudaMemcpyHostToDevice);    
  
  // 矩阵 B --> 矩阵 d_B
  Matrix d_B;    
  d_B.width = d_B.stride = B.width; 
  d_B.height = B.height;   
  size = B.width * B.height * sizeof(float);
  cudaMalloc(&d_B.elements, size);    
  cudaMemcpy(d_B.elements, B.elements, size, cudaMemcpyHostToDevice);    

  // Allocate C in device memory   
  // 为矩阵 C 分配空间
  Matrix d_C;    
  d_C.width = d_C.stride = C.width; 
  d_C.height = C.height;    
  size = C.width * C.height * sizeof(float);   
  cudaMalloc(&d_C.elements, size);    
  
  // Invoke kernel    
  // 启动核函数
  dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);    
  dim3 dimGrid(B.width / dimBlock.x, A.height / dimBlock.y);    
  MatMulKernel<<<dimGrid, dimBlock>>>(d_A, d_B, d_C);    

  // Read C from device memory
  // 将计算得到的矩阵 C 的数据 从 device memory 送回 host memory    
  cudaMemcpy(C.elements, d_C.elements, size, cudaMemcpyDeviceToHost);  

  // Free device memory    
  // 释放 设备内存
  cudaFree(d_A.elements);    
  cudaFree(d_B.elements);    
  cudaFree(d_C.elements); 
} 

// Matrix multiplication kernel called by MatMul() 
__global__ void MatMulKernel(Matrix A, Matrix B, Matrix C) 
{   
  // Block row and column
  // 行索引 和 列索引    
  int blockRow = blockIdx.y;    
  int blockCol = blockIdx.x;    
  
  // Each thread block computes one sub-matrix Csub of C    
  Matrix Csub = GetSubMatrix(C, blockRow, blockCol);   

  // Each thread computes one element of Csub    
  // by accumulating results into Cvalue    
  float Cvalue = 0;    

  // Thread row and column within Csub    
  int row = threadIdx.y;    
  int col = threadIdx.x;   
   // Loop over all the sub-matrices of A and B that are    
   // required to compute Csub    
   // Multiply each pair of sub-matrices together   
    // and accumulate the results    
    for (int m = 0; m < (A.width / BLOCK_SIZE); ++m) {       
      // Get sub-matrix Asub of A        
      Matrix Asub = GetSubMatrix(A, blockRow, m);      

      // Get sub-matrix Bsub of B        
      Matrix Bsub = GetSubMatrix(B, m, blockCol);       

      // Shared memory used to store Asub and Bsub respectively        
      __shared__ float As[BLOCK_SIZE][BLOCK_SIZE];        
      __shared__ float Bs[BLOCK_SIZE][BLOCK_SIZE];        

      // Load Asub and Bsub from device memory to shared memory      
      // Each thread loads one element of each sub-matrix        
      As[row][col] = GetElement(Asub, row, col);       
      Bs[row][col] = GetElement(Bsub, row, col);        
        
      // Synchronize to make sure the sub-matrices are loaded       
      // before starting the computation        
      __syncthreads();

      // Multiply Asub and Bsub together        
      for (int e = 0; e < BLOCK_SIZE; ++e)            
        Cvalue += As[row][e] * Bs[e][col];        
            
        // Synchronize to make sure that the preceding        
        // computation is done before loading two new       
        // sub-matrices of A and B in the next iteration       
      __syncthreads();    
    }    
      // Write Csub to device memory    
      // Each thread writes one element    
      
    SetElement(Csub, row, col, Cvalue); 
}